package intl;
/*

Copyright 2001,2003,2006  The Apache Software Foundation

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

*/
import java.io.*;
import java.nio.charset.Charset;
import java.text.AttributedCharacterIterator;
import java.util.*;
import org.apache.fop.fonts.*;



/**
* Handles the processing of arabic text. In particular it determines the
* form each arabic char should take. It also contains methods for substituting
* plain arabic glyphs with their shaped forms. This is needed when the arabic
* text is rendered using an AWT font.
*
* @author <a href="mailto:bella.robinson@cmis.csiro.au">Bella Robinson</a>
* @version $Id: ArabicTextHandler.java 397765 2006-04-28 05:31:11Z cam $
*/
public class ArabicText {

 private final static int arabicStart = 0x0600;
 private final static int arabicEnd = 0x06FF;

 private static final AttributedCharacterIterator.Attribute ARABIC_FORM =
     TextAttribute.ARABIC_FORM;
 private static final Integer ARABIC_NONE =
     TextAttribute.ARABIC_NONE;
 private static final Integer ARABIC_ISOLATED =
     TextAttribute.ARABIC_ISOLATED;
 private static final Integer ARABIC_FINAL =
     TextAttribute.ARABIC_TERMINAL;
 private static final Integer ARABIC_INITIAL =
     TextAttribute.ARABIC_INITIAL;
 private static final Integer ARABIC_MEDIAL =
     TextAttribute.ARABIC_MEDIAL;
 
 private static final Integer RIGHT_JOINING = new Integer(0x5);
 private static final Integer DUAL_JOINING = new Integer(0x6);
 private static final Integer NON_JOINING = new Integer(0x7);
 
 private static Hashtable forms = null;
 private static Hashtable shapings = null;


 /**
  * private ctor prevents unnecessary instantiation of this class.
  */
 private ArabicText() {}
 
 /**
  * Returns true if the char is a standard arabic char.
  * (ie. within the range U+0600 - U+6FF)
  *
  * @param c The character to test.
  * @return True if the char is arabic, false otherwise.
  */
 public static boolean arabicChar(char c) {
     if (c >= arabicStart && c <= arabicEnd)
         return true;
     return false;
 }
 
 public static boolean arabicChar(char c, boolean anyForm) {
	 if (arabicChar(c) || (anyForm &&
			 (c >= 0xFE70 && c <= 0xFEFF		// Arabic Presentation Forms-B
			 || c >= 0xFB50 && c <= 0xFDFF		// Arabic Presentation Forms-A
			 || c >= 0x0750 && c <= 0x077F)))	// Arabic Supplement
		 return true;
	 else return false;
 }

 /**
  * Returns true if the char is transparent.
  *
  * @param c The character to test.
  * @return True if the character is transparent, false otherwise.
  */
 public static boolean arabicCharTransparent(char c) {
     int charVal = c;
     if ((charVal  < 0x064B) || (charVal > 0x06ED))
         return false;

     if ((charVal <= 0x0655)                      ||
         (charVal == 0x0670)                      ||
         (charVal >= 0x06D6 && charVal <= 0x06E4) ||
         (charVal >= 0x06E7 && charVal <= 0x06E8) ||
         (charVal >= 0x06EA)) {
         return true;
     }
     return false;
 }

 /**
  * Returns true if the character shapes to the right. Note that duel
  * shaping characters also shape to the right and so will return true.
  *
  * @param c The character to test.
  * @return True if the character shapes to the right, false otherwise.
  */
 private static boolean arabicCharShapesRight(char c) {
	 String code = Integer.toHexString(c);
	 if (code.length() < 4) code = "0" + code;
	 Integer sh = (Integer)getShapings().get(code.toUpperCase());
	 if (sh != null && (sh == RIGHT_JOINING || sh == DUAL_JOINING)) return true;
	 return false;	 
 }

 /**
  * Returns true if character has duel shaping.
  *
  * @param c The character to test.
  * @return True if the character is duel shaping, false otherwise.
  */
 private static boolean arabicCharShapesDuel(char c) {
	 String code = Integer.toHexString(c);
	 if (code.length() < 4) code = "0" + code;
	 Integer sh = (Integer)getShapings().get(code.toUpperCase());
	 if (sh != null && sh == DUAL_JOINING) return true;
	 return false;	 
 }

 /**
  * Returns true if character shapes to the left. Note that duel
  * shaping characters also shape to the left and so will return true.
  *
  * @param c The character to test.
  * @return True if the character shapes to the left, false otherwise.
  */
 private static boolean arabicCharShapesLeft(char c) {
     return arabicCharShapesDuel(c);
 }
 
 public static boolean hasSubstitute(char ch1, char ch2, Font fs) {
	 return (getSubstituteChar(ch1, ch2, -1, fs) != -1);
 }

 /**
  * Will try and find a substitute character of the specified form.
  *
  * @param ch1 The first character of two to replace.
  * @param ch2 The second character of two to replace.
  * @param form Indicates the required arabic form.
  * (isolated = 1, final = 2, initial = 3, medial = 4)
  *
  * @return The unicode value of the substutute char, or -1 if no susbtitue
  * exists.
  */
  
 public static int getSubstituteChar(char ch1, char ch2, int form, Font fs) {
	 return getSubstituteChar((int)ch1, (int)ch2, form, fs);
 }
 
 public static int getSubstituteChar(int ch1, int ch2, int form, Font fs) {
	 
	 // if form == -1 return true if the character has any substitution.
	 String low = Integer.toHexString(ch1);
	 if (low.length() < 4) low = "0" + low;
	 if (ch2 != -1) {
		 String low2 = Integer.toHexString(ch2);
		 if (low2.length() < 4) low2 = "0" + low2;
		 low = low + " " + low2;
	 }
	 	 
	 ArrayList al = (ArrayList)getForms().get(low.toUpperCase());
	 if (al == null) return -1;
	 if (form == -1) return 0;
	 
	 for (int i = 0; i < al.size(); i++) {
		 int[] inh = (int[])al.get(i);
		 if (inh[1] == form) {
			 return (fs.hasChar((char)inh[0])) ? inh[0] : -1;
		 }
	 }
	 
	 return -1;
 }
 
 private static Hashtable getForms() {
	 if (forms == null) loadForms();
	 return forms;
 }
 
 public static int getSubstituteChar(char ch, int form, Font fs) {
	 return getSubstituteChar((int)ch, -1, form, fs);
 }
 
 private static void loadForms() {
	 forms = new Hashtable();	 
	 try {
		InputStream is = ClassLoader.getSystemResourceAsStream("forms.txt");
		if (is == null) {
			System.out.println("Cannot find forms.txt file.");
			return;
		}
		InputStreamReader isr = new InputStreamReader(is, Charset.forName("UTF-16"));
		if (isr == null) {
			System.out.println("Cannot find forms.txt file.");
			return;
		}
		BufferedReader br = new BufferedReader(isr);
		// BufferedReader br = new BufferedReader(new InputStreamReader(new FileInputStream("C:/fop/forms.txt"), Charset.forName("UTF-16")));
		String s, comment, high, low;
		StringTokenizer t;
		int location = -1;
		while ((s = br.readLine()) != null) {
			if (s.trim().startsWith("#") || s.trim().length() == 0)
				continue;
			t = new StringTokenizer(s, ";");
			high = t.nextToken().trim();
			t.nextToken();
			t.nextToken();
			//t.nextToken();
			low = t.nextToken().trim();
			
			comment = s.substring(s.lastIndexOf(")"));
			if (comment.indexOf("ISOLATED") != -1) location = ARABIC_ISOLATED.intValue();
			else if (comment.indexOf("FINAL") != -1) location = ARABIC_FINAL.intValue();
			else if (comment.indexOf("INITIAL") != -1) location = ARABIC_INITIAL.intValue();
			else if (comment.indexOf("MEDIAL") != -1) location = ARABIC_MEDIAL.intValue();
		
			if (forms.get(low) == null) {
				ArrayList rep = new ArrayList(1);
				rep.add(new int[]{Integer.parseInt(high, 16), location});
				forms.put(low, rep);
			} else {
				((ArrayList) forms.get(low)).add(new int[]{Integer.parseInt(high, 16), location});
				
			}
		}
		br.close();
		isr.close();
		is.close();
	} catch (Exception e) {
		e.printStackTrace();
	}
 }
 
 private static Hashtable getShapings() {
	 if (shapings == null) loadShapings();
	 return shapings;
 }
 
 private static void loadShapings() {
	shapings = new Hashtable();	 	
	try {
		InputStream is = ClassLoader.getSystemResourceAsStream("shapings.txt");
		if (is == null) {
			System.out.println("Cannot find forms.txt file.");
			return;
		}
		InputStreamReader isr = new InputStreamReader(is);
		if (isr == null) {
			System.out.println("Cannot find forms.txt file.");
			return;
		}
		BufferedReader br = new BufferedReader(isr);
		// BufferedReader br = new BufferedReader(new InputStreamReader(new FileInputStream("C:/fop/shapings.txt")));
		String s, code, shaping;
		StringTokenizer t;
		while ((s = br.readLine()) != null) {
			t = new StringTokenizer(s, ";");
			code = t.nextToken().trim();
			t.nextToken();
			shaping = t.nextToken();
			
			if (shaping.indexOf("R") != -1) shapings.put(code, RIGHT_JOINING);
			else if (shaping.indexOf("D") != -1 || shaping.indexOf("C") != -1 || shaping.indexOf("L") != -1) shapings.put(code, DUAL_JOINING);
			else if (shaping.indexOf("U") != -1) shapings.put(code, NON_JOINING);
		}
		br.close();
		isr.close();
		is.close();
	 } catch (Exception e) {
		 e.printStackTrace();
	 }
 }
 public static boolean containsArabic(String source) {
	 return containsArabic(source, false);
 }
 
 public static boolean containsArabic(String source, boolean anyForm) {
	 return containsArabic(source.toCharArray(), anyForm);
 }
 
 public static boolean containsArabic(char[] c, boolean anyForm) {
	 for (int i = c.length-1; i >= 0; i--)
		 if (arabicChar(c[i], true)) return true;
	 return false;
 }
 
 public static boolean containsArabic(char[] c) {
	 return containsArabic(c, false);
 }
 
public static int[] findLigature(String source, int pos, Font fs) {
	 char[] c = source.toCharArray();
	 int nextCharPos = (c.length-pos < 16) ? c.length : pos + 4;
	 String key = null;
	 for (int i=pos; i < nextCharPos; i++) {
		 String low = Integer.toHexString(c[i]);
		 while (low.length() < 4) low = "0" + low;
		 if (key == null) key = low;
		 else key = key + " " + low;
	 }
	 key = key.toUpperCase();
	 
	 for (;(nextCharPos > pos && (key.lastIndexOf(" ") != -1)); nextCharPos--) {
		 ArrayList al = (ArrayList)getForms().get(key);
		 if (al != null) {
			 Integer form = getCharForm(source, pos, nextCharPos - pos);
			 for (int i = 0; i < al.size(); i++) {
				 int[] inh = (int[])al.get(i);
				 if (inh[1] == form.intValue() && fs.hasChar((char)inh[0])) {
					 return new int[]{nextCharPos - pos, inh[0]};
				 }
			 }
		 }
		 key = key.substring(0, key.lastIndexOf(" "));
	 }	 
	 return new int[]{0, 0};
 }
 
 public static Integer getCharForm(String source, int pos, int length) {
	 char[] c = source.toCharArray();
	 if (arabicCharTransparent(c[pos])) return ARABIC_NONE;
	 
	 int prevCharPos = pos, nextCharPos = pos+length;
	 while (--prevCharPos >= 0)
		 if (!arabicCharTransparent(c[prevCharPos])) break;
	 while (nextCharPos < c.length) {
		 if (!arabicCharTransparent(c[nextCharPos])) break;
		 nextCharPos++;
	 }
	 
	 boolean prevSL = (prevCharPos >= 0) ? arabicCharShapesLeft(c[prevCharPos]) : false,
			 nextSR = (nextCharPos < c.length) ? arabicCharShapesRight((c[nextCharPos])) : false;
			 
	if (prevSL && nextSR) return ARABIC_MEDIAL;
	if (prevSL && !nextSR) return ARABIC_FINAL;
	if (!prevSL && nextSR) return ARABIC_INITIAL;
	 
	return ARABIC_ISOLATED;
 }
 
 public static String createArabic(String source, Font fs) {
	 StringBuilder target = new StringBuilder();
	 char[] c = source.toCharArray();
	 for (int pos=0; pos < c.length; pos++) {
		 if (arabicChar(c[pos]) || c[pos] == ' ') {
			 int[] ret = findLigature(source, pos, fs);
			 int ligatureLength = ret[0];
			 char ligature = (char)ret[1];
			 if (ligatureLength > 0) {
				 target.append(ligature);
				 pos += ligatureLength-1;
			 } else {
				 int subst = getSubstituteChar(c[pos], getCharForm(source, pos, 1).intValue(), fs);
				 if (subst != -1)
					 target.append((char)subst);
				 else
					 target.append(c[pos]);
			 }
		 } else target.append(c[pos]);
	 }
	 return target.toString();
 }
}